/*
 * asm.c --
 *
 *	Los procedimientos de este fichero realizan el ensamblado
 *	y el desensamblado de las instrucciones del DLXV.
 *
 * Copyright 1989 Regents of the University of California
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  The University of California
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 */

/****************************************************************
 * PROYECTO FINAL DE CARRERA:		        		*
 *	Simulador Vectorial para PC bajo entorno Windows	*
 * 		                                		*
 *      Autor: Roberto Calpe Lorenzo				*
 *                                                              *
 ****************************************************************/

#include <bwcc.h>
#include <alloc.h>
#include <ctype.h>
#include <errno.h>
#include "hash.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tcl.h"
#include "dlx.h"
#include "asm.h"
#include "gp.h"
#include "sym.h"

/*
 * La siguiente estructura encapsula el estado de la carga de uno o
 * mas ficheros de ensamblador, tambien es usada para la comunicacion
 * entre los procedimientos que realizan el ensamblado.
 */

typedef struct {
    char *file;			/* Nombre del fichero que esta siendo
				 * ensamblado. */
    long int lineNum;		/* Numero de linea del fichero(-1 si la
				 * informacion que esta siendo ensamblada
				 * no proviene de un fichero ). */
    char *line;			/* Contenido de la linea que esta siendo
				 * ensamblada. */
    unsigned long int codeAddr;	/* Siguiente direccion del area de codigo
				 * en la que poner informacion. */
    unsigned long int dataAddr;	/* Siguiente direccion del area de datos en
				 * la que poner informacion. */
    unsigned long int dot;	/* Direccion en la que poner la informacion
				 * (puede ser codigo o datos). */

    /*
     * Informacion usada para construir una concatenacion de los mensajes
     * de error que ocurren mientras se lee el fichero.
     */

    char huge *end;		/* Direccion del byte NULL al final del
				 * mensaje; aadir nuevos mensajes aqui */
    long int totalBytes;	/* # de bytes ocupados por el mensaje.
				 * 0 significa que aun no ha ocurrido ningun
				 * error */
    long int errorCount;	/* Si ocurren demasiados errores hay que
				 * abandonar. */
    long int flags;		/* Varios flags: Ver abajo los valores. */

    char huge *message;		/* Puntero al mensaje actual. NULL significa
				 * que no ha ocurrido un error durante la
				 * actual pasada. */

} LoadInfo;

/*
 * Flags para la estructura LoadInfo:
 *
 * ASM_CODE -			1 significa que estamos ensamblando en el
 *				area de codigo;  0 significa que estamos
 *				ensamblando en el area de datos
 * ASM_SIZE_ONLY -		1 significa que este es el primer paso, donde
 *				la unica cosa importante es el tamao (suprime
 * 				todos los mensajes de error).
 * ASM_ALIGN_0 -		1 significa que un comando "align 0" esta en
 *				efecto.
 */

#define ASM_CODE	1L
#define ASM_SIZE_ONLY	2L
#define ASM_ALIGN_0	4L

#define ASM_MAX_ERRORS	20L

/*
 * Los #defines siguientes especifican las diferentes clase de instrucciones,
 * tal y como estan definidas en las paginas D-4 a D-6 del libro de Kane.
 * Ademas han sido aadidas las clases de operaciones correspondientes a la
 * implementacion del DLXV.
 * Estas clases son usadas durante el ensamblado, e indican los diferentes
 * formatos que pueden ser tomados por los operandos para un determinado
 * codigo de operacion.
 *
 * NO_ARGS -		sin operandos
 * LOAD -		(registro, direccion)
 * STORE -		(direccion, registro)
 * LUI -		(dest, expresion de 16-bit)
 * ARITH -		(dest, src1, sr2c) o (dest/src1, src2)
 *			o (dest, src1, 16-bit inmediato)
 *			o (dest/src1, 16-bit inmediato)
 * MULDIV -		igual que ARITH (subconjunto especial para manejar
 *			instrucciones mult/divide/rem, ya que operan sobre
 *			registros en coma flotante)
 * BRANCH_0_OP -	(etiqueta) el registro fuente es implicito
 * BRANCH_1_OP -	(src1, etiqueta)
 * JUMP -		(etiqueta) o (src1)
 * SRC1 -		(src1)
 * LABEL -		(etiqueta)
 * MOVE -		(dest,src1)
 * FLOAD -		como LOAD, para simple/doble precision
 * FSTORE -		como STORE, para simple/doble precision
 * FARITH -		como ARITH, para simple/doble precision
 * FCOMPARE -		compara dos registros en coma flotante
 * TRAP -		(numero de trap)
 * VLOAD		como LOAD, para operaciones vectoriales
 * VSTORE		como STORE, para operaciones vectoriales
 * VARITH		como ARITH, para operaciones vectoriales
 * VCOMPARE             comparar dos registros vectoriales
 * VMOVE		(dest, src1) con los registros especiales VLR y VM
 * VMASK		(regV, src1) creacion del vector de indices
 * VPOP                 (dest) almacena la cuenta de unos del VM
 * VSLOAD               (destV, src1, src2) carga con stride en src2
 * VSSTORE		(dest, src1, src2V) almacena con stride en src1
 * VSILOAD		(destV, src1, src2V) carga con vector de indices
 * VSISTORE		(dest, src1V, src2V) almacena con vector de indices
 */

#define NO_ARGS		0L
#define LOAD		1L
#define STORE		2L
#define LUI		3L
#define ARITH		4L
#define MULDIV		5L
#define BRANCH_0_OP	7L
#define BRANCH_1_OP	8L
#define JUMP		9L
#define SRC1		10L
#define LABEL		11L
#define MOVE		12L
#define FLOAD		13L
#define FSTORE		14L
#define FARITH		15L
#define FCOMPARE	16L
#define TRAP		17L
#define VLOAD		18L
#define VSTORE		19L
#define VARITH		20L
#define VCOMPARE	21L
#define VMOVE		22L
#define VMASK		23L
#define	VPOP		24L
#define VSLOAD		25L
#define VSSTORE		26L
#define VSILOAD		27L
#define VSISTORE	28L


/*
 * La siguiente tabla proporciona el maximo y el minimo # de argumentos
 * permitidos para cada clase especificada arriba.
 */

static int far minArgs[] =
    {0, 3, 3, 2, 3, 3, 3, 1, 2, 1, 1, 1, 2, 3, 3, 3, 2, 1, 3, 3, 3, 2, 2, 2, 2, 3, 3, 3, 3};
static int far maxArgs[] =
    {0, 3, 3, 2, 3, 3, 3, 1, 2, 1, 1, 1, 2, 3, 3, 3, 2, 1, 3, 3, 3, 2, 2, 2, 2, 3, 3, 3, 3};

/*
 * La siguiente estructura es usada durante el ensamblado y el desensamblado.
 * Para cada Codigo de operacion definido existe una estructura de este tipo.
 */

typedef struct {
    char *name;			/* Nombre del codigo de operacion,
				 * ej:"add". */
    long int class;		/* Clase de instruccion (ver tabla de
				 * arriba). */
    long int op;		/* Patron de bits correspondiente a
				 * esta instruccion. */
    long int mask;		/* Usado para desensamblar: si estos bits
				 * coinciden con op, entonces hay que usar
				 * este opcode para desensamblar. 0 significa
				 * que es una instruccion sintetizada que no
				 * exste en forma nativa, por tanto debe ser
				 * ignorado durante el desensamblado. */
    long int other;		/* Este campo es usado cuando el ensamblador
				 * genera varias instrucciones para un solo
				 * opcode, o cuando instrucciones diferentes
				 * pueden ser generadas para el mismo opcode
				 * (ej: add -> addi). El significado de este
				 * campo depende de la clase; ver codigo en
				 * Asm_Assemble. */
    long int flags;		/* Combinacion de bits mediante la operacion OR
				 * proporcionando informaciom varia durante el
				 * ensamblado, tal com el chequeo de rango. Ver
				 * abajo los valores. */
    long int rangeMask;		/* Mascara para usar en el chequeo de rango:
				 * para valores sin signo, niguno de estos bits
				 * debe estar activado. Para comprobar valores
				 * con signo o todos o ninguno deben estar
				 * activados. */
} OpcodeInfo;

/*
 * Bits para el campo "flags" , usados en la comprobacion de rango:
 *
 * CHECK_FIRST -		1 significa comprobar el primer argumento,
 *	 			por si es un valor inmediato.
 * CHECK_LAST -			1 significa comprobar el ultimo argumento,
 *				por si es un valor inmediato.
 * CHECK_NEXT_TO_LAST -		1 significa comprobar el penultimo argumento,
 *				por si es inmediato.
 * IMMEDIATE_REQ -		1 significa que el argumento ha de ser un
 *				valor inmediato.
 * SIGN_EXTENDED -		1 significa que el valor inmediato es con signo
 * FIRST_F
 * SECOND_F
 * THIRD_F			El 1/2/3  registro deben ser de simple
 *				precision.
 * FIRST_V
 * SECOND_v
 * THIRD_V			El 1/2/3 registro deben ser vectoriales.
 *
 * FIRST_D
 * SECOND_D
 * THIRD_D			El 1/2/3 registro deben ser de doble
 *				precision.
 *
 * FIRST_VLR
 * SECOND_VLR			El 1/2 registro deben ser el vlr
 *
 * FIRST_VM
 * SECOND_VM			El 1/2 registro deben ser el vm
 */

#define CHECK_LAST		0x1L
#define CHECK_NEXT_TO_LAST	0x2L
#define IMMEDIATE_REQ		0x4L
#define SIGN_EXTENDED		0x8L
#define CHECK_FIRST		0x10L
#define FIRST_VEC		0x20L
#define SECOND_VEC		0x40L
#define THIRD_VEC		0x80L
#define FIRST_F			0x100L
#define SECOND_F		0x200L
#define THIRD_F			0x400L
#define FIRST_D			0x800L
#define SECOND_D		0x1000L
#define THIRD_D			0x2000L
#define FIRST_VLR		0x4000L
#define SECOND_VLR		0x8000L
#define FIRST_VM		0x10000L
#define SECOND_VM		0x20000L


/*
 * Tabla de todas las instrucciones conocidas:
 */

OpcodeInfo far opcodes[] = {
    {"add", ARITH, 0x20, 0xfc00003fL, 0x20000000L,
	    CHECK_LAST|SIGN_EXTENDED, 0xffff8000L},
    {"addd", FARITH, 0x04000004L, 0xfc00003fL, 0, FIRST_D|SECOND_D|THIRD_D, 0},
    {"addf", FARITH, 0x04000000L, 0xfc00003fL, 0, FIRST_F|SECOND_F|THIRD_F, 0},
    {"addi", ARITH, 0x20000000L, 0xfc000000L, 0,
	    CHECK_LAST|SIGN_EXTENDED|IMMEDIATE_REQ, 0xffff8000L},
    {"addsv", VARITH, 0x0400001FL, 0xfc00003fL, 0, FIRST_VEC | SECOND_D | THIRD_VEC, 0},
    {"addu", ARITH, 0x21, 0xfc00003fL, 0x24000000L, CHECK_LAST, 0xffff0000L},
    {"addui", ARITH, 0x24000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"addv", VARITH, 0x0400001EL, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC | THIRD_VEC, 0},
    {"and", ARITH, 0x24, 0xfc00003fL, 0x30000000L, CHECK_LAST, 0xffff0000L},
    {"andi", ARITH, 0x30000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"beqz", BRANCH_1_OP, 0x10000000L, 0xfc000000L, 0, 0, 0},
    {"bfpf", BRANCH_0_OP, 0x18000000L, 0xfc000000L, 0, 0, 0},
    {"bfpt", BRANCH_0_OP, 0x1c000000L, 0xfc000000L, 0, 0, 0},
    {"bnez", BRANCH_1_OP, 0x14000000L, 0xfc000000L, 0, 0, 0},
    {"cvi", VMASK, 0x0000003a, 0xfc00003fL, 0, FIRST_VEC, 0},
    {"cvm", NO_ARGS, 0x00000039, 0xfc00003fL, 0, 0, 0},
    {"cvtd2f", MOVE, 0x0400000aL, 0xfc00003fL, 0, FIRST_F|SECOND_D, 0},
    {"cvtd2i", MOVE, 0x0400000bL, 0xfc00003fL, 0, FIRST_F|SECOND_D, 0},
    {"cvtf2d", MOVE, 0x04000008L, 0xfc00003fL, 0, FIRST_D|SECOND_F, 0},
    {"cvtf2i", MOVE, 0x04000009L, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"cvti2d", MOVE, 0x0400000dL, 0xfc00003fL, 0, FIRST_D|SECOND_F, 0},
    {"cvti2f", MOVE, 0x0400000cL, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"div", MULDIV, 0x0400000fL, 0xfc00003fL, 0, FIRST_F|SECOND_F|THIRD_F, 0},
    {"divd", FARITH, 0x04000007L, 0xfc00003fL, 0, FIRST_D|SECOND_D|THIRD_D, 0},
    {"divf", FARITH, 0x04000003L, 0xfc00003fL, 0, FIRST_F|SECOND_F|THIRD_F, 0},
    {"divsv", VARITH, 0x04000027L, 0xfc00003fL, 0, FIRST_VEC | SECOND_D | THIRD_VEC, 0},
    {"divu", MULDIV, 0x04000017L, 0xfc00003fL, 0, FIRST_F|SECOND_F|THIRD_F, 0},
    {"divv", VARITH, 0x04000025L, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC | THIRD_VEC, 0},
    {"divvs", VARITH, 0x04000026L, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC | THIRD_D, 0},
    {"eqd", FCOMPARE, 0x04000018L, 0xfc00003fL, 0, FIRST_D|SECOND_D, 0},
    {"eqf", FCOMPARE, 0x04000010L, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"ged", FCOMPARE, 0x0400001dL, 0xfc00003fL, 0, FIRST_D|SECOND_D, 0},
    {"gef", FCOMPARE, 0x04000015L, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"gtd", FCOMPARE, 0x0400001bL, 0xfc00003fL, 0, FIRST_D|SECOND_D, 0},
    {"gtf", FCOMPARE, 0x04000013L, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"j", JUMP, 0x08000000L, 0xfc000000L, 0x48000000L, 0, 0},
    {"jal", JUMP, 0x0c000000L, 0xfc000000L, 0x4c000000L, 0, 0},
    {"jalr", SRC1, 0x4c000000L, 0xfc000000L, 0, 0, 0},
    {"jr", SRC1, 0x48000000L, 0xfc000000L, 0, 0, 0},
    {"lb", LOAD, 0x80000000L, 0xfc000000L, 0,
	    CHECK_NEXT_TO_LAST|IMMEDIATE_REQ|SIGN_EXTENDED, 0xffff8000L},
    {"lbu", LOAD, 0x90000000L, 0xfc000000L, 0,
	    CHECK_NEXT_TO_LAST|IMMEDIATE_REQ|SIGN_EXTENDED, 0xffff8000L},
    {"ld", FLOAD, 0x9c000000L, 0xfc000000L, 0,
	    CHECK_NEXT_TO_LAST|IMMEDIATE_REQ|SIGN_EXTENDED|FIRST_D, 0xffff8000L},
    {"led", FCOMPARE, 0x0400001cL, 0xfc00003fL, 0, FIRST_D|SECOND_D, 0},
    {"lef", FCOMPARE, 0x04000014L, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"lf", FLOAD, 0x98000000L, 0xfc000000L, 0,
	    CHECK_NEXT_TO_LAST|IMMEDIATE_REQ|SIGN_EXTENDED|FIRST_F, 0xffff8000L},
    {"lh", LOAD, 0x84000000L, 0xfc000000L, 0,
	    CHECK_NEXT_TO_LAST|IMMEDIATE_REQ|SIGN_EXTENDED, 0xffff8000L},
    {"lhi", LUI, 0x3c000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"lhu", LOAD, 0x94000000L, 0xfc000000L, 0,
	    CHECK_NEXT_TO_LAST|IMMEDIATE_REQ|SIGN_EXTENDED, 0xffff8000L},
    {"ltd", FCOMPARE, 0x0400001aL, 0xfc00003fL, 0, FIRST_D|SECOND_D, 0},
    {"ltf", FCOMPARE, 0x04000012L, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"lv", VLOAD, 0x88000000L, 0xfc000000L, 0,
	    CHECK_NEXT_TO_LAST | IMMEDIATE_REQ | SIGN_EXTENDED | FIRST_VEC, 0xffff8000L},
    {"lvi", VSILOAD, 0x0000003D, 0xfc00003fL, 0, FIRST_VEC | THIRD_VEC, 0},
    {"lvws", VSLOAD, 0x0000003B, 0xfc00003fL, 0, FIRST_VEC, 0},
    {"lw", LOAD, 0x8c000000L, 0xfc000000L, 0,
	    CHECK_NEXT_TO_LAST|IMMEDIATE_REQ|SIGN_EXTENDED, 0xffff8000L},
    {"movd", MOVE, 0x00000033, 0xfc00003fL, 0, FIRST_D|SECOND_D, 0},
    {"movf", MOVE, 0x00000032, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"movf2s", VMOVE, 0x36, 0xfc00003fL, 0, FIRST_VM | SECOND_D, 0},
    {"movfp2i", MOVE, 0x00000034, 0xfc00003fL, 0, SECOND_F, 0},
    {"movi2fp", MOVE, 0x00000035, 0xfc00003fL, 0, FIRST_F, 0},
    {"movi2s", VMOVE, 0x00000030, 0xfc00003fL, 0, FIRST_VLR, 0},
    {"movs2f", VMOVE, 0x37, 0xfc00003fL, 0, FIRST_D | SECOND_VM, 0},
    {"movs2i", VMOVE, 0x00000031, 0xfc00003fL, 0, SECOND_VLR, 0},
    {"mult", MULDIV, 0x0400000eL, 0xfc00003fL, 0, FIRST_F|SECOND_F|THIRD_F, 0},
    {"multd", FARITH, 0x04000006L, 0xfc00003fL, 0, FIRST_D|SECOND_D|THIRD_D, 0},
    {"multf", FARITH, 0x04000002L, 0xfc00003fL, 0, FIRST_F|SECOND_F|THIRD_F, 0},
    {"multsv", VARITH, 0x04000024L, 0xfc00003fL, 0, FIRST_VEC | SECOND_D | THIRD_VEC, 0},
    {"multu", MULDIV, 0x04000016L, 0xfc00003fL, 0, FIRST_F|SECOND_F|THIRD_F, 0},
    {"multv", VARITH, 0x04000023L, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC | THIRD_VEC, 0},
    {"ned", FCOMPARE, 0x04000019L, 0xfc00003fL, 0, FIRST_D|SECOND_D, 0},
    {"nef", FCOMPARE, 0x04000011L, 0xfc00003fL, 0, FIRST_F|SECOND_F, 0},
    {"nop", NO_ARGS, 0x0, 0xffffffffL, 0, 0, 0},
    {"or", ARITH, 0x25, 0xfc00003fL, 0x34000000L, CHECK_LAST, 0xffff0000L},
    {"ori", ARITH, 0x34000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"pop", VPOP, 0x00000038, 0xfc00003fL, 0, SECOND_VM, 0},
    {"rfe", LABEL, 0x40000000L, 0xfc000000L, 0, 0, 0xfe000000L},
    {"sb", STORE, 0xa0000000L, 0xfc000000L, 0,
	    CHECK_FIRST|IMMEDIATE_REQ|SIGN_EXTENDED, 0xffff8000L},
    {"sd", FSTORE, 0xbc000000L, 0xfc000000L, 0,
	    CHECK_FIRST|IMMEDIATE_REQ|SIGN_EXTENDED|THIRD_D, 0xffff8000L},
    {"seq", ARITH, 0x00000028, 0xfc00003fL, 0x60000000L,
	    CHECK_LAST|SIGN_EXTENDED, 0xffff8000L},
    {"seqi", ARITH, 0x60000000L, 0xfc000000L, 0,
	    CHECK_LAST|SIGN_EXTENDED|IMMEDIATE_REQ, 0xffff8000L},
    {"seqsv", VCOMPARE, 0x0400002EL, 0xfc00003fL, 0, FIRST_D | SECOND_VEC, 0},
    {"sequ", ARITH, 0x00000010, 0xfc00003fL, 0xc0000000L,
	    CHECK_LAST, 0xffff0000L},
    {"sequi", ARITH, 0xc0000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"seqv", VCOMPARE, 0x04000028L, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC, 0},
    {"sf", FSTORE, 0xb8000000L, 0xfc000000L, 0,
	    CHECK_FIRST|IMMEDIATE_REQ|SIGN_EXTENDED|THIRD_F, 0xffff8000L},
    {"sge", ARITH, 0x0000002d, 0xfc00003fL, 0x74000000L,
	    CHECK_LAST|SIGN_EXTENDED, 0xffff8000L},
    {"sgei", ARITH, 0x74000000L, 0xfc000000L, 0,
	    CHECK_LAST|SIGN_EXTENDED|IMMEDIATE_REQ, 0xffff8000L},
    {"sgesv", VCOMPARE, 0x04000032L, 0xfc00003fL, 0, FIRST_D | SECOND_VEC, 0},
    {"sgeu", ARITH, 0x00000015, 0xfc00003fL, 0xd4000000L,
	    CHECK_LAST, 0xffff0000L},
    {"sgeui", ARITH, 0xd4000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"sgev", VCOMPARE, 0x0400002CL, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC, 0},
    {"sgt", ARITH, 0x0000002b, 0xfc00003fL, 0x6c000000L,
	    CHECK_LAST|SIGN_EXTENDED, 0xffff8000L},
    {"sgti", ARITH, 0x6c000000L, 0xfc000000L, 0,
	    CHECK_LAST|SIGN_EXTENDED|IMMEDIATE_REQ, 0xffff8000L},
    {"sgtsv", VCOMPARE, 0x04000030L, 0xfc00003fL, 0, FIRST_D | SECOND_VEC, 0},
    {"sgtu", ARITH, 0x00000013, 0xfc00003fL, 0xcc000000L,
	    CHECK_LAST, 0xffff0000L},
    {"sgtui", ARITH, 0xcc000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"sgtv", VCOMPARE, 0x0400002AL, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC, 0},
    {"sh", STORE, 0xa4000000L, 0xfc000000L, 0,
	    CHECK_FIRST|IMMEDIATE_REQ|SIGN_EXTENDED, 0xffff8000L},
    {"sle", ARITH, 0x0000002c, 0xfc00003fL, 0x70000000L,
	    CHECK_LAST|SIGN_EXTENDED, 0xffff8000L},
    {"slei", ARITH, 0x70000000L, 0xfc000000L, 0,
	    CHECK_LAST|SIGN_EXTENDED|IMMEDIATE_REQ, 0xffff8000L},
    {"slesv", VCOMPARE, 0x04000033L, 0xfc00003fL, 0, FIRST_D | SECOND_VEC, 0},
    {"sleu", ARITH, 0x00000014, 0xfc00003fL, 0xd0000000L,
	    CHECK_LAST, 0xffff0000L},
    {"sleui", ARITH, 0xd0000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"slev", VCOMPARE, 0x0400002DL, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC, 0},
    {"sll", ARITH, 0x4, 0xfc00003fL, 0x50000000L, CHECK_LAST, 0xffffffe0L},
    {"slli", ARITH, 0x50000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"slt", ARITH, 0x0000002a, 0xfc00003fL, 0x68000000L,
	    CHECK_LAST|SIGN_EXTENDED, 0xffff8000L},
    {"slti", ARITH, 0x68000000L, 0xfc000000L, 0,
	    CHECK_LAST|SIGN_EXTENDED|IMMEDIATE_REQ, 0xffff8000L},
    {"sltsv", VCOMPARE, 0x04000031L, 0xfc00003fL, 0, FIRST_D | SECOND_VEC, 0},
    {"sltu", ARITH, 0x00000012, 0xfc00003fL, 0xc8000000L,
	    CHECK_LAST, 0xffff0000L},
    {"sltui", ARITH, 0xc8000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"sltv", VCOMPARE, 0x0400002BL, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC, 0},
    {"sne", ARITH, 0x00000029, 0xfc00003fL, 0x64000000L,
	    CHECK_LAST|SIGN_EXTENDED, 0xffff8000L},
    {"snei", ARITH, 0x64000000L, 0xfc000000L, 0,
	    CHECK_LAST|SIGN_EXTENDED|IMMEDIATE_REQ, 0xffff8000L},
    {"snesv", VCOMPARE, 0x0400002FL, 0xfc00003fL, 0, FIRST_D | SECOND_VEC, 0},
    {"sneu", ARITH, 0x00000011, 0xfc00003fL, 0xc4000000L,
	    CHECK_LAST, 0xffff0000L},
    {"sneui", ARITH, 0xc4000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"snev", VCOMPARE, 0x04000029L, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC, 0},
    {"sra", ARITH, 0x7, 0xfc00003fL, 0x5c000000L, CHECK_LAST, 0xffffffe0L},
    {"srai", ARITH, 0x5c000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"srl", ARITH, 0x6, 0xfc00003fL, 0x58000000L, CHECK_LAST, 0xffffffe0L},
    {"srli", ARITH, 0x58000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"sub", ARITH, 0x22, 0xfc00003fL, 0x28000000L,
	    CHECK_LAST|SIGN_EXTENDED, 0xffff8000L},
    {"subd", FARITH, 0x04000005L, 0xfc00003fL, 0, FIRST_D|SECOND_D|THIRD_D, 0},
    {"subf", FARITH, 0x04000001L, 0xfc00003fL, 0, FIRST_F|SECOND_F|THIRD_F, 0},
    {"subi", ARITH, 0x28000000L, 0xfc000000L, 0,
	    CHECK_LAST|SIGN_EXTENDED|IMMEDIATE_REQ, 0xffff8000L},
    {"subu", ARITH, 0x23, 0xfc00003fL, 0x2c000000L, CHECK_LAST, 0xffff0000L},
    {"subui", ARITH, 0x2c000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {"subv", VARITH, 0x04000020L, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC | THIRD_VEC, 0},
    {"subvs", VARITH, 0x04000021L, 0xfc00003fL, 0, FIRST_VEC | SECOND_VEC | THIRD_D, 0},
    {"subsv", VARITH, 0x04000022L, 0xfc00003fL, 0, FIRST_VEC | SECOND_D | THIRD_VEC, 0},
    {"sv", VSTORE, 0xA8000000L, 0xfc000000L, 0,
	    CHECK_FIRST | IMMEDIATE_REQ | SIGN_EXTENDED | THIRD_VEC, 0xffff8000L},
    {"svi", VSISTORE, 0x0000003E, 0xfc00003fL, 0, SECOND_VEC | THIRD_VEC, 0},
    {"svws", VSSTORE, 0x0000003C, 0xfc00003fL, 0, THIRD_VEC, 0},
    {"sw", STORE, 0xac000000L, 0xfc000000L, 0,
	    CHECK_FIRST|IMMEDIATE_REQ|SIGN_EXTENDED, 0xffff8000L},
    {"trap", TRAP, 0x44000000L, 0xfc000000L, 0,
	    CHECK_FIRST|IMMEDIATE_REQ, 0xfc000000L},
    {"xor", ARITH, 0x26, 0xfc00003fL, 0x38000000L, CHECK_LAST, 0xffff0000L},
    {"xori", ARITH, 0x38000000L, 0xfc000000L, 0,
	    CHECK_LAST|IMMEDIATE_REQ, 0xffff0000L},
    {NULL, NO_ARGS, 0, 0, 0, 0, 0}
};

/*
 * Valores de codigos de operacion que son utilizados en este modulo:
 */

#define SUBU_OP			0x23L
#define ADDI_OP			0x20000000L
#define ADDIU_OP		0x24000000L
#define LUI_OP			0x3c000000L
#define ORI_OP			0x34000000L
#define BEQ_OP			0x10000000L
#define BNE_OP			0x14000000L
#define LOAD_IMM(reg, x)	(0x20000000L | ((reg) << 16) | ((x) & 0xffffL))
#define LOAD_IMM_UNS(reg, x)	(0x34000000L | ((reg) << 16) | ((x) & 0xffffL))

/*
 * Tabla para mapear el numero del registro con su nombre correspondiente:
 */

char * far Asm_RegNames[] = {
    "r0", "r1", "r2", "r3",
    "r4", "r5", "r6", "r7",
    "r8", "r9", "r10", "r11",
    "r12", "r13", "r14", "r15",
    "r16", "r17", "r18", "r19",
    "r20", "r21", "r22", "r23",
    "r24", "r25", "r26", "r27",
    "r28", "r29", "r30", "r31",
    "f0", "f1", "f2", "f3",
    "f4", "f5", "f6", "f7",
    "f8", "f9", "f10", "f11",
    "f12", "f13", "f14", "f15",
    "f16", "f17", "f18", "f19",
    "f20", "f21", "f22", "f23",
    "f24", "f25", "f26", "f27",
    "f28", "f29", "f30", "f31",
    "pc", "npc", "vlr", "vm", "vmlo",
    "v0", "v1", "v2", "v3", "v4", "v5",
    "v6", "v7", "hi", "lo",	/* Algunos de estos pueden ser eliminados */
    "status", "cause", "epc"
};

/*
 * Tamao de la instruccion mas larga que podemos ensamblar, en palabras:
 */

#define ASM_MAX_WORDS 5L		/* probablemente 1 */

/*
 * Declaraciones de procedimientos definidos en este fichero:
 */

static void	AddErrMsg();
static void	IndicateError();
static void	ReadFile();
static long int	StoreWords();

/*
 *----------------------------------------------------------------------
 *
 * Asm_Assemble --
 *
 *	Dado un string con una instruccion en ensamblador, devuelve
 *	el codigo binario correspondiente a la instruccion.
 *
 * Resultados:
 *	El valor devuelto es un OK (normalmente TCL_OK mas
 *	un string vacio).  Si el ensamblado se ha completado
 *	satisfactoriamente, entonces en *sizePtr se almacena el # de
 *	palabras ensambladas (pueden ser mas de 1 para pseudo-
 *	instrucciones especiales), y la(s) palabra(s) en *codePtr se
 *	rellena con la instruccion actual.
 *
 * Efectos laterales:
 *	Ninguno.
 *
 *----------------------------------------------------------------------
 */

long int
Asm_Assemble(machPtr, fileName, string, dot, wholeLine, sizeOnly,
	sizePtr, codePtr)
    DLXV *machPtr;			/* Maquina para la cual se realiza el
					 * ensamblado:
					 * usado para mantener la tabla de
					 * simbolos y para comunicar
					 * errores. */
    char *fileName;			/* Nombre de fichero a utilizar para la
					 * comprobacion de simbolos. Ver el
					 * procedimiento GetSym en sym.c para
					 * mas informacion. */
    char *string;			/* Instruccion en ensamblador del
					 * DLXV. */
    unsigned long int dot;		/* En que posicion de memoria sera
					 * almacenada la instruccion. */
    char *wholeLine;			/* Linea completa conteniendo el string;
					 * usado al imprimir los mensajes de
					 * error. */
    long int sizeOnly;			/* No-cero significa que esta es la
					 * primera pasada, por tanto se
					 * ignoran los errores y solo se
					 * calcula el tamao de la instruccion
					 */
    long int *sizePtr;			/* Aqui se almacena el # de palabras
					 * ensambladas para esta instruccion. */
    long int *codePtr;			/* Aqui se almacena el codigo
					 * que ha sido ensamblado. */
{
    register OpcodeInfo *insPtr;	/* Informacion acerca de la
					 * instruccion. */
    register char *p;			/* Caracter del string que en cada
					 * momento esta siendo procesado. */
    char *errMsg;
    char *opStart;
    int length;
    char msg[100];
    long int isReg[3];			/* Dice si cada operando de la
					 * instruccion es un registro. */
    long int operands[3];		/* Valor de cada operando (# de reg,
					 * inmediato, cantidad a desplazar,
					 * etc.). */
    char *argStart[3];			/* Primeros caracteres de los
					 * argumentos (para emision
					 * de errores). */
    int numOps;				/* Numero de operandos en la
					 * instruccion. */
    long int done;
    long int requireF, requireD;	/* Indica si el operando actual debe
					 * ser un registro flotante de simple
					 * o doble precision. */
    long int requireV;			/* Indica si el operando actual debe
					 * ser un registro vectorial */
    long int requireVM, requireVLR;	/* Indica si el operando actual debe
					 * ser el registro "vm" o el "vlr" */
    /*
     * Encuentra el nombre de la instruccion, y lo busca en la tabla.
     */
    *sizePtr = 1;
    for (p = string; (*p == ' ') || (*p == '\t'); p++) {
	/* Bucle sin cuerpo. */
    }
    opStart = p;
    for ( ; isalnum(*p); p++) {
	/* Bucle sin cuerpo. */
    }
    length = (int)(p-opStart);
    if (length > 0) {
	for (insPtr = opcodes; insPtr->name != NULL; insPtr++) {
	    if ((insPtr->name[0] == opStart[0])
		    && (_fstrncmp(insPtr->name, opStart, length) == 0)
		    && (insPtr->name[length] == 0)) {
		codePtr[0] = insPtr->op;
		goto gotIns;
	    }
	}
    }
    errMsg = "Unknown operation code";
/*    p =  opStart; */
    goto error;

    /*
     * Analiza los tres campos de operandos en la instruccion, almacenando
     * la informacion en isReg[], operands[], y numOps.
     */

    gotIns:
    isReg[0] = isReg[1] = isReg[2] = 0;
    operands[0] = operands[1] = operands[2] = 0;
    for (numOps = 0; numOps < 3; numOps++) {
	char *end, savedChar;
	long int result;

      /*
       * Encuentra el caracter de comienzo para este especificador de operando.
       */

	while ((*p == ' ') || (*p == '\t')) {
	    p++;
	}

	argStart[numOps] = p;

	/*
	 * El siguiente codigo es un caso especial para manejar
	 * instrucciones del tipo load-store. Acepta una expresion
	 * opcional seguida por un registro opcional entre parentesis.
	 */

	if (((numOps == 0) && ((insPtr->class == STORE) || (insPtr->class == VSTORE) ||
	 (insPtr->class == FSTORE))) || ((numOps == 1) && ((insPtr->class == LOAD) ||
	 (insPtr->class == VLOAD) || (insPtr->class == FLOAD)))) {
	    if (*p == '(') {
		operands[numOps] = 0;
	    } else {
		if (Sym_EvalExpr(machPtr, fileName, p, sizeOnly,
			&operands[numOps], &end) != TCL_OK) {
		    errMsg = (char far *)machPtr->result;
		    goto error;
		}
		p = end;
	    }
	    while ((*p == ' ') || (*p == '\t')) {
		p++;
	    }
	    argStart[++numOps] = p;
	    isReg[numOps] = 1;
	    operands[numOps] = 0;
	    if (*p == '(') {
		for (p++; *p != ')'; p++) {
		    if ((*p == 0)  || (*p == ';')) {
			p = argStart[numOps];
			errMsg = "Error ) after register base";
			goto error;
		    }
		}
		savedChar = *p;
		*p = 0;
		result = Sym_GetSym(machPtr, fileName, argStart[numOps]+1,
			SYM_REGS_OK, (unsigned long int *) &operands[numOps]);
		*p = savedChar;
		if (result != SYM_REGISTER) {
		    p = argStart[numOps]+1;
		    errMsg = (result == SYM_FREG_FOUND)
				? "Floating point register not valid"
				: "Incorrect name";
		    goto error;
		}
		p++;
	    }
	/* Lee hasta que encuentra un separador */
	    for (done = 0; !done; ) {
		switch (*p) {
		case ',' :
		case '#' :
		case ';' :
		case '\n' :
		case '\0' :
		    done = 1;
		    break;
		case ' ' :
		case '\t' :
		    break;
		default :
		    errMsg = "Unrecognized characters";
		    goto error;
		}
		if (!done)
		    p++;
	    }

	    end = p;
	    goto about_to_continue;
	}

	/*
	 * Volvemos al caso normal.  Hemos de encontrar el final del
	 * especificador de operando actual.
	 */

	while ((*p != ',') && (*p != ';') && (*p != 0) && (*p != '\n')) {
	    p++;
	}
	end = p;
	if (p == argStart[numOps]) {
	    if (numOps == 0) {
		break;
	    }
	    errMsg = "Operand specification is empty";
	    goto error;
	}
	for (p--; (*p == ' ') || (*p == '\t'); p--) {
	    /* Bucle sin cuerpo; Unicamente vuelve hacia atras sobre los
	     * espacios encontrados */
	}
	p++;

	/*
	 * Ahora buscamos el tipo del operando.
	 */

	savedChar = *p;
	*p = 0;

	/* determina si necesitamos un registro simple/doble precision ahora */

	requireF = ((numOps == 0) && (insPtr->flags & FIRST_F)) ||
		   ((numOps == 1) && (insPtr->flags & SECOND_F)) ||
		   ((numOps == 2) && (insPtr->flags & THIRD_F));
	requireD = ((numOps == 0) && (insPtr->flags & FIRST_D)) ||
		   ((numOps == 1) && (insPtr->flags & SECOND_D)) ||
		   ((numOps == 2) && (insPtr->flags & THIRD_D));

	/* determina si necesitamos un registro vectorial ahora */

	requireV = ((numOps == 0) && (insPtr->flags & FIRST_VEC)) ||
		   ((numOps == 1) && (insPtr->flags & SECOND_VEC)) ||
		   ((numOps == 2) && (insPtr->flags & THIRD_VEC));

	/* determina si necesitamos el registro "vm" */

	requireVM = ((numOps == 0) && (insPtr->flags & FIRST_VM)) ||
		   ((numOps == 1) && (insPtr->flags & SECOND_VM));

	/* determina si necesitamos el registro "vlr" */

	requireVLR = ((numOps == 0) && (insPtr->flags & FIRST_VLR)) ||
		   ((numOps == 1) && (insPtr->flags & SECOND_VLR));

	if (requireVM){
	    result = Sym_GetSym(machPtr, fileName, argStart[numOps],
		SYM_PSEUDO_OK, (unsigned long int *) &operands[numOps]);
	    if ((result != SYM_REGISTER) || (operands[numOps] != (NEXT_PC_REG +2))){
		*p = savedChar;
		p = argStart[numOps];
		errMsg = "Register required \"vm\"";
		goto error;
		}
		isReg[numOps] = 1;
	    }
	else{ if (requireVLR){
	    result = Sym_GetSym(machPtr, fileName, argStart[numOps],
		SYM_PSEUDO_OK, (unsigned long int *) &operands[numOps]);
	    if ( (result != SYM_REGISTER) || (operands[numOps] != (NEXT_PC_REG +1))){
		*p = savedChar;
		p = argStart[numOps];
		errMsg = "Register required \"vlr\"";
		goto error;
		}
		isReg[numOps] = 1;
	     }
	else{
	if (requireV){
	    result = Sym_GetSym(machPtr, fileName, argStart[numOps],
		SYM_VREGS_OK, (unsigned long int *) &operands[numOps]);
	    if (result != SYM_REGISTER) {
		*p = savedChar;
		p = argStart[numOps];
		errMsg = "Vector register required";
		goto error;
	    }
	    isReg[numOps] = 1;
	}
	else { if (requireF || requireD) {
	    result = Sym_GetSym(machPtr, fileName, argStart[numOps],
		SYM_FREGS_OK, (unsigned long int *) &operands[numOps]);
	    if (result != SYM_REGISTER) {
		*p = savedChar;
		p = argStart[numOps];
		errMsg = "Floating point register required";
		goto error;
	    }
	    isReg[numOps] = 1;
	    if (requireD && (operands[numOps] & 1)) {
		*p = savedChar;
		p = argStart[numOps];
		errMsg = "Floating point register required, double precision";
		goto error;
	    }
	} else {
	    result = Sym_GetSym(machPtr, fileName, argStart[numOps],
		SYM_REGS_OK, (unsigned long int *) &operands[numOps]);
	    if (result == SYM_REGISTER) {
		isReg[numOps] = 1;
	    } else if (result == SYM_FREG_FOUND) {
		*p = savedChar;
		p = argStart[numOps];
		errMsg = "Floating point register not valid";
		goto error;
	    } else if (result == SYM_VREG_FOUND) {
		*p = savedChar;
		p = argStart[numOps];
		errMsg = "Vector register not valid";
		goto error;
	    } else if (result != SYM_FOUND) {
		char *term;

		if (Sym_EvalExpr(machPtr, (char far *) NULL, argStart[numOps],
			sizeOnly, (long int *) &operands[numOps], &term) != TCL_OK) {
		    *p = savedChar;
		    p = argStart[numOps];
		    errMsg = "Unknown operand specification";
		    goto error;
		}
		if (*term != 0) {
		    *p = savedChar;
		    p = term;
		    errMsg = "Unrecognizes characters";
		    goto error;
		}
	    }
	}
	}
	}
	}
	*p = savedChar;

	/*
	 * Miramos si es el ultimo argumento.  Si no es asi, nos saltamos
	 * la coma de separacion.
	 */

about_to_continue:
	p = end;
	if (*p != ',') {
	    numOps++;
	    break;
	}
	if (numOps == 2) {
	    errMsg = "Three operands required";
	    goto error;
	}
	p++;
    }

    /*
     * Comprueba el numero de argumentos.
     */

    if ((numOps < minArgs[(unsigned int)insPtr->class])
	    || (numOps > maxArgs[(unsigned int)insPtr->class])) {
	if (minArgs[(unsigned int)insPtr->class] == maxArgs[(unsigned int)insPtr->class]) {
	    sprintf(msg, "Incorrect # operand (must be %d)",
		    minArgs[(unsigned int)insPtr->class]);
	} else {
	    sprintf(msg, "Incorrect # operand (must be %d o %d)",
		    minArgs[(unsigned int)insPtr->class], maxArgs[(unsigned int)insPtr->class]);
	}
	p = argStart[0];
	errMsg = msg;
	goto error;
    }

    /*
     * Comprueba los operandos inmediatos para que tengan el rango adecuado.
     */

    if (insPtr->flags & (CHECK_LAST | CHECK_NEXT_TO_LAST | CHECK_FIRST)) {
	int i;

	if (insPtr->flags & CHECK_LAST) {
	    i = numOps-1;
	} else if (insPtr->flags & CHECK_FIRST) {
	    i = 0;
	} else {
	    i = numOps-2;
	}
	if (i >= 0) {
	    if (isReg[i]) {
		if (insPtr->flags & IMMEDIATE_REQ) {
		    p = argStart[i];
		    regIllegal:
		    errMsg = "This register type not permited";
		    goto error;
		}
	    } else {
		long int j;
		j = operands[i] & insPtr->rangeMask;
		if (j != 0) {
		    if (!(insPtr->flags & SIGN_EXTENDED)
			    || (j != insPtr->rangeMask)) {
			p = argStart[i];
			sprintf(msg, "Direct operand 0x%lx out of range",
				operands[i]);
			errMsg = msg;
			goto error;
		    }
		}
	    }
	}
    }

    /*
     * Por ultimo maneja todos los casos segun el tipo de instruccion.
     */

    *sizePtr = 1;
    switch (insPtr->class) {
	case NO_ARGS:
	    codePtr[0] = insPtr->op;
	    break;

	case LOAD:
	case FLOAD:
	case VLOAD:
	    codePtr[0] = insPtr->op | (operands[0] << 16)
		    | (operands[1] & 0xffff) | (operands[2] << 21);
	    break;

	case STORE:
	case FSTORE:
	case VSTORE:
	    codePtr[0] = insPtr->op | (operands[0] & 0xffff)
		    | (operands[1] << 21) | (operands[2] << 16);
	    break;

	case LUI:
	    codePtr[0] = insPtr->op | (operands[0] << 16)
		    | (operands[1] & 0xffff);
	    break;

	/*
	 * La clase principal de instrucciones aritmeticas pueden ensamblarse
	 * de varias maneras. Muchas instrucciones pueden terminar usando el
	 * codigo de operacion registro-registro o con un codigo de operacion
	 * inmediato, el cual es almacenado en insPtr->other. Si la instruccion
	 * debe usar unicamente la forma inmediata, un valor especial de
	 * insPtr->other indica este hecho.
	 */

	case ARITH:
	case FARITH:
	case MULDIV:
	case VARITH:
	    if (!isReg[0]) {
		p = argStart[0];
		regRequired:
		errMsg = "Operand must be register";
		goto error;
	    } else if (!isReg[1]) {
		p = argStart[1];
		goto regRequired;
	    }
	    if (insPtr->class == ARITH) {
		if (isReg[2]) {
		    codePtr[0] = insPtr->op | (operands[0] << 11)
			    | (operands[1] << 21) | (operands[2] << 16);
		} else if (insPtr->flags & IMMEDIATE_REQ) {
		    codePtr[0] = insPtr->op | (operands[0] << 16)
			    | (operands[1] << 21) | (operands[2] & 0xffff);
		} else {
		    codePtr[0] = insPtr->other | (operands[0] << 16)
			    | (operands[1] << 21) | (operands[2] & 0xffff);
		}
	    } else if ((insPtr->class == MULDIV) || (insPtr->class == FARITH) ||
	     (insPtr->class == VARITH)) {
		if (!isReg[2]) {
		    p = argStart[2];
		    goto regRequired;
		}
		codePtr[0] = insPtr->op | (operands[0] << 11)
			| (operands[1] << 21) | (operands[2] << 16);
	    }
	    break;

	/*
	 * Saltos:  Genera (y comprueba) el desplazamiento del salto, lo cual
	 * se hace igual para todas las clases de salto. Por ultimo maneja las
	 * diferentes subclases de forma diferente.
	 */

	case JUMP:
	    if (isReg[0])	/* Se trata como SRC1 */
		codePtr[0] = insPtr->other | (operands[0] << 21);
		/* Este codigo no para aqui, esto se hace para que el usuario
		 * pueda utilizar j en lugar de jr */
	case BRANCH_0_OP:
	case BRANCH_1_OP:
	case LABEL: {
	    long int disp, mask;

	    if (isReg[numOps-1]) {
		p = argStart[numOps-1];
		goto regIllegal;
	    }
	    disp = operands[numOps-1];
	    if (disp & 0x3) {
		p = argStart[numOps-1];
		errMsg = "Jump destination is not aligned (word-aligned)";
		goto error;
	    }
	    disp = disp - (dot+4);
	    if ((insPtr->class == JUMP) || (insPtr->class == LABEL))
		mask = 0xfe000000L;
	    else
		mask = 0xffff8000L;
	    if ((disp & mask) && ((disp & mask) != mask)) {
		badDisp:
		p = argStart[numOps-1];
		sprintf(msg, "Jump destination too far (offset 0x%lx)", disp);
		errMsg = msg;
		goto error;
	    }
	    if (insPtr->class == BRANCH_1_OP) {
		codePtr[0] = insPtr->op | (disp & 0xffff) | (operands[0] << 21);
	    } else if (insPtr->class == BRANCH_0_OP) {
		codePtr[0] = insPtr->op | (disp & 0xffff);
	    } else {  /* JUMP or LABEL */
		codePtr[0] = insPtr->op | (disp & 0x3ffffffL);
	    }
	    break;
	}

	case TRAP:
	    if (isReg[0]) {
		p = argStart[0];
		goto regIllegal;
	    }
	    codePtr[0] = insPtr->op | (operands[0] & 0x3ffffffL);
	    break;

	case SRC1:
	    if (!isReg[0]) {
		p = argStart[0];
		goto regRequired;
	    }
	    codePtr[0] = insPtr->op | (operands[0] << 21);
	    break;

	case MOVE:
		/* Estas son instrucciones del tipo Registro */
	    if (!isReg[0]) {
		p = argStart[0];
		goto regRequired;
	    }
	    if (!isReg[1]) {
		p = argStart[1];
		goto regRequired;
	    }
	    codePtr[0] = insPtr->op | (operands[0] << 11) | (operands[1] << 21);
	    break;

	case FCOMPARE:
	case VCOMPARE:
		/* Estas son instrucciones del tipo Registro */
	    if (!isReg[0]) {
		p = argStart[0];
		goto regRequired;
	    }
	    if (!isReg[1]) {
		p = argStart[1];
		goto regRequired;
	    }
	    codePtr[0] = insPtr->op | (operands[0] << 21) | (operands[1] << 16);
	    break;

	case VMOVE:
	    if ((!isReg[0]) || (!isReg[1])){
		p = argStart[0];
		goto regRequired;
	    }
	    if (insPtr->flags & FIRST_VLR){
		operands[0] = 1;}
	    if (insPtr->flags & SECOND_VLR){
		operands[1] = 1;}
	    if (insPtr->flags & FIRST_VM){
		operands[0] = 2;}
	    if (insPtr->flags & SECOND_VM){
		operands[1] = 2;}
	    codePtr[0] = insPtr->op | (operands[0] << 11) | (operands[1] << 21);
	    break;

	case VPOP:
	    if (!isReg[0]) {
		p = argStart[0];
		goto regRequired;
	    }
	    if (insPtr->flags & SECOND_VM){
		operands[1] = 2;}
	    codePtr[0] = insPtr->op | (operands[0] << 11) | (operands[1] << 21);
	    break;

	case VMASK:
	    if (!isReg[1]) {
		p = argStart[0];
		goto regRequired;
	    }
	    codePtr[0] = insPtr->op | (operands[0] << 11) | (operands[1] << 21);
	    break;

	case VSLOAD:
	case VSILOAD:
	    if (!isReg[1]) {
		p = argStart[0];
		goto regRequired;
	    }
	    if (!isReg[2]) {
		p = argStart[0];
		goto regRequired;
	    }
	    codePtr[0] = insPtr->op | (operands[0] << 11) | (operands[1] << 21) | (operands[2] << 16);
	    break;

	case VSSTORE:
	case VSISTORE:
	    if (!isReg[0]) {
		p = argStart[0];
		goto regRequired;
	    }
	    if (!isReg[1]) {
		p = argStart[0];
		goto regRequired;
	    }
	    codePtr[0] = insPtr->op | (operands[0] << 11) | (operands[1] << 21) | (operands[2] << 16);
	    break;

	default:
	    errMsg = "Internal error:  Unknown instruction";
	    goto error;
    }

    /*
     * Nos aseguramos de que no quede basura en la linea despues de haber
     * ensamblado la instruccion.
     */

    while (isspace(*p)) {
	p++;
    }
    if ((*p != 0) && (*p != ';')) {
	errMsg = "Characters after end of line";
	goto error;
    }
    return TCL_OK;

    /*
     * Genera un mensaje de error lo suficientemente comprensible.
     */

    error:
    IndicateError(machPtr, errMsg, wholeLine, p);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * Asm_Disassemble --
 *
 *	Dada una instruccion, devuelve un string que describe la instruccion
 *	en formato ensamblador.
 *
 * Resultados:
 *	El valor devuelto es un string, El cual describe la instruccion o
 *	contiene un mensaje diciendo que no se reconocio la instruccion. El
 *	string es alojado en memoria estaticamente, lo que significa que
 *	cambiara en la siguiente llamada al procedimiento.
 *
 * Efectos Laterales:
 *	Ninguno.
 *
 *----------------------------------------------------------------------
 */

char *
Asm_Disassemble(machPtr, ins, pc)
    DLXV *machPtr;		/* Maquina a utilizar para la tabla de
				 * simbolos. */
    long int ins;		/* Contenido de la instruccion. */
    unsigned long int pc;	/* Direccion de memoria en la cual la
				 * instruccion esta almacenada. */
{
    register OpcodeInfo *opPtr;
    OpcodeInfo *bestPtr;
    static char string[200];
    long int field, bestMask;
    long int reg1, reg2;

    /*
     * Igualamos la instruccion con nuestra tabla de instrucciones para
     * conocer que instruccion es:
     */

    for (bestMask = 0, opPtr = opcodes; opPtr->name != NULL; opPtr++) {
	if (opPtr->mask == 0) {
	    continue;
	}
	if ((ins & opPtr->mask) != (opPtr->op)) {
	    continue;
	}
	if ((bestMask & opPtr->mask) != opPtr->mask) {
	    bestMask = opPtr->mask;
	    bestPtr = opPtr;
	}
    }
    if (bestMask == 0) {
	sprintf(string, "Instruction not recognized (0x%lx)", ins);
	return string;
    }
    opPtr = bestPtr;

    /*
     * Tratamos la instruccion segun su tipo.
     */

    switch (opPtr->class) {
	case NO_ARGS:
	    sprintf(string, "%s", opPtr->name);
	    break;

	case LOAD:
	    field = ins & 0xffff;
	    sprintf(string, "%s %s,%s(%s)", opPtr->name,
		    Asm_RegNames[(unsigned int)((ins >> 16) & 0x1f)],
		    Sym_GetString(machPtr, (unsigned long int) field),
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)]);
	    break;

	case FLOAD:
	    field = ins & 0xffff;
	    sprintf(string, "%s %s,%s(%s)", opPtr->name,
		    Asm_RegNames[(unsigned int)(((ins >> 16) & 0x1f) + 32)],
		    Sym_GetString(machPtr, (unsigned long int) field),
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)]);
	    break;

	case VLOAD:
	    field = ins & 0xffff;
	    sprintf(string, "%s %s,%s(%s)", opPtr->name,
		    Asm_RegNames[(unsigned int)(((ins >> 16) & 0x1f) + TOTAL_REGS)],
		    Sym_GetString(machPtr, (unsigned long int) field),
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)]);
	    break;

	case STORE:
	    field = ins & 0xffff;
	    sprintf(string, "%s %s(%s),%s", opPtr->name,
		    Sym_GetString(machPtr, (unsigned long int) field),
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)],
		    Asm_RegNames[(unsigned int)((ins >> 16) & 0x1f)]);
	    break;

	case FSTORE:
	    field = ins & 0xffff;
	    sprintf(string, "%s %s(%s),%s", opPtr->name,
		    Sym_GetString(machPtr, (unsigned long int) field),
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)],
		    Asm_RegNames[(unsigned int)(((ins >> 16) & 0x1f) + 32)]);
	    break;

	case VSTORE:
	    field = ins & 0xffff;
	    sprintf(string, "%s %s(%s),%s", opPtr->name,
		    Sym_GetString(machPtr, (unsigned long int) field),
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)],
		    Asm_RegNames[(unsigned int)(((ins >> 16) & 0x1f) + TOTAL_REGS)]);
	    break;

	case LUI:
	    field = ins & 0xffff;
	    sprintf(string, "%s %s,0x%lx", opPtr->name,
		    Asm_RegNames[(unsigned int)((ins >> 16) & 0x1f)], field);
	    break;

	case ARITH:
	    if (opPtr->flags & IMMEDIATE_REQ) {
		field = ins & 0xffff;
		sprintf(string, "%s %s,%s,0x%lx", opPtr->name,
			Asm_RegNames[(unsigned int)((ins >> 16) & 0x1f)],
			Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)], field);
	    } else {
		sprintf(string, "%s %s,%s,%s", opPtr->name,
			Asm_RegNames[(unsigned int)((ins >> 11) & 0x1f)],
			Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)],
			Asm_RegNames[(unsigned int)((ins >> 16) & 0x1f)]);
	    }
	    break;

	case FARITH:
	case MULDIV:
	    sprintf(string, "%s %s,%s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)(((ins >> 11) & 0x1f) + 32)],
		    Asm_RegNames[(unsigned int)(((ins >> 21) & 0x1f) + 32)],
		    Asm_RegNames[(unsigned int)(((ins >> 16) & 0x1f) + 32)]);
	    break;

	case VARITH:
	    reg1 = (ins >> 21) & 0x1f;
	    reg2 = (ins >> 16) & 0x1f;
	    if (opPtr->flags & SECOND_D)
		{reg1 += 32;}
		else { reg1 += TOTAL_REGS;}
	    if (opPtr->flags & THIRD_D)
		{reg2 += 32;}
		else { reg2 += TOTAL_REGS;}
	    sprintf(string, "%s %s,%s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)(((ins >> 11) & 0x1f) + TOTAL_REGS)],
		    Asm_RegNames[(unsigned int)reg1],
		    Asm_RegNames[(unsigned int)reg2]);
	    break;

	case BRANCH_0_OP:
	    field = (ins & 0xffff);
	    if (field & 0x8000) {
		field |= 0xffff0000L;
	    }
	    field += pc + 4;
	    sprintf(string, "%s %s", opPtr->name,
		    Sym_GetString(machPtr, (unsigned long int) field));
	    break;

	case BRANCH_1_OP:
	    field = (ins & 0xffff);
	    if (field & 0x8000) {
		field |= 0xffff0000L;
	    }
	    field += pc + 4;
	    sprintf(string, "%s %s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)],
		    Sym_GetString(machPtr, (unsigned long int) field));
	    break;

	case TRAP:
	    field = (ins & 0x3ffffffL);
	    sprintf(string, "%s 0x%lx", opPtr->name, field);
	    break;

	case LABEL:
	case JUMP:
	    field = (ins & 0x3ffffffL);
	    if (field & 0x2000000L) {
		field |= 0xfc000000L;
	    }
	    field += pc + 4;
	    sprintf(string, "%s %s", opPtr->name,
		    Sym_GetString(machPtr, (unsigned long int) field));
	    break;

	case SRC1:
	    sprintf(string, "%s %s", opPtr->name,
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)]);
	    break;

	case MOVE:
	    reg1 = (ins >> 11) & 0x1f;
	    reg2 = (ins >> 21) & 0x1f;
	    if ((opPtr->flags & FIRST_F) || (opPtr->flags & FIRST_D))
		reg1 += 32;
	    if ((opPtr->flags & SECOND_F) || (opPtr->flags & SECOND_D))
		reg2 += 32;
	    sprintf(string, "%s %s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)reg1],
		    Asm_RegNames[(unsigned int)reg2]);
	    break;
	case FCOMPARE:
		/* Esta es una instruccion de tipo registro */
	    reg1 = (ins >> 21) & 0x1f;
	    reg2 = (ins >> 16) & 0x1f;
	    if ((opPtr->flags & FIRST_F) || (opPtr->flags & FIRST_D))
		reg1 += 32;
	    if ((opPtr->flags & SECOND_F) || (opPtr->flags & SECOND_D))
		reg2 += 32;
	    sprintf(string, "%s %s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)reg1],
		    Asm_RegNames[(unsigned int)reg2]);
	    break;

	case VCOMPARE:
	    reg1 = (ins >> 21) & 0x1f;
	    if (opPtr->flags & FIRST_D)
		{reg1 += 32;}
		else { reg1 += TOTAL_REGS;}
	    sprintf(string, "%s %s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)reg1],
		    Asm_RegNames[(unsigned int)(((ins >> 16) & 0x1f) + TOTAL_REGS)]);
	    break;

	case VMOVE:
	    reg1 = (ins >> 11) & 0x1f;
	    reg2 = (ins >> 21) & 0x1f;
	    if (opPtr->flags & FIRST_VLR){
		reg1 += NEXT_PC_REG;}
	    else {
	    if (opPtr->flags & SECOND_VLR){
		reg2 += NEXT_PC_REG;}
	    else {
	    if (opPtr->flags & FIRST_VM){
		reg1 += NEXT_PC_REG;}
	    if (opPtr->flags & SECOND_VM){
		reg2 += NEXT_PC_REG;}
	    if (opPtr->flags & FIRST_D){
		reg1 += 32;}
	    if (opPtr->flags & SECOND_D){
		reg2 += 32;}
	    }
	    }
	    sprintf(string, "%s %s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)reg1],
		    Asm_RegNames[(unsigned int)reg2]);
	    break;

	case VPOP:
	    sprintf(string, "%s %s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)((ins >> 11) & 0x1f)],
		    Asm_RegNames[(unsigned int)(((ins >> 21) & 0x1f) + NEXT_PC_REG)]);
	    break;

	case VMASK:
	    sprintf(string, "%s %s,%s", opPtr->name,
		    Asm_RegNames[(unsigned int)(((ins >> 11) & 0x1f) + TOTAL_REGS)],
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)]);
	    break;

	case VSLOAD:
	    reg2 = (ins >> 16) & 0x1f;
	    if (opPtr->flags & THIRD_VEC){
		reg2 += TOTAL_REGS;}
	    sprintf(string, "%s %s,(%s,%s)", opPtr->name,
		    Asm_RegNames[(unsigned int)(((ins >> 11) & 0x1f) + TOTAL_REGS)],
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)],
		    Asm_RegNames[(unsigned int)reg2]);
	    break;

	case VSILOAD:
	    reg2 = (ins >> 16) & 0x1f;
	    if (opPtr->flags & THIRD_VEC){
		reg2 += TOTAL_REGS;}
	    sprintf(string, "%s %s,(%s+%s)", opPtr->name,
		    Asm_RegNames[(unsigned int)(((ins >> 11) & 0x1f) + TOTAL_REGS)],
		    Asm_RegNames[(unsigned int)((ins >> 21) & 0x1f)],
		    Asm_RegNames[(unsigned int)reg2]);
	    break;

	case VSSTORE:
	    reg2 = (ins >> 21) & 0x1f;
	    if (opPtr->flags & SECOND_VEC){
		reg2 += TOTAL_REGS;}
	    sprintf(string, "%s (%s,%s),%s", opPtr->name,
		    Asm_RegNames[(unsigned int)((ins >> 11) & 0x1f)],
		    Asm_RegNames[(unsigned int)reg2],
		    Asm_RegNames[(unsigned int)(((ins >> 16) & 0x1f) + TOTAL_REGS)]);
	    break;

	case VSISTORE:
	    reg2 = (ins >> 21) & 0x1f;
	    if (opPtr->flags & SECOND_VEC){
		reg2 += TOTAL_REGS;}
	    sprintf(string, "%s (%s+%s),%s", opPtr->name,
		    Asm_RegNames[(unsigned int)((ins >> 11) & 0x1f)],
		    Asm_RegNames[(unsigned int)reg2],
		    Asm_RegNames[(unsigned int)(((ins >> 16) & 0x1f) + TOTAL_REGS)]);
	    break;

	default:
	    sprintf(string, "Disassembled instruction confused (0x%lx)", ins);
	    break;
    }
    return string;
}

/*
 *----------------------------------------------------------------------
 *
 * Asm_AsmCmd --
 *
 *	Este procedimiento es utilizado para procesar el comando "Ensamblar"
 *	Ver la documentacion de usuario para saber lo que hace dicho comando.
 *
 * Resultados:
 *	Un resultado de OK o ERROR.
 *
 * Efectos Laterales:
 *	Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

/* Se utilizan argumentos */
long int
Asm_AsmCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;			/* Descripcion de la maquina. */
    int numarg;				/* Numero de argumentos. */
    char argumentos[argus][largus];	/* Cadena de argumentos. */
{
    long int result;
    unsigned long int pc;
    char *end;
    long int size, code[ASM_MAX_WORDS];
    extern char far salidas[199];

    if ((numarg != 2) && (numarg != 3)) {
      wsprintf(salidas,
		"Incorrect # argument:  must be \"%.50s\" instruccion [pc]",
		argumentos[0]);
      BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
      machPtr->result[0] = '\0';
      return TCL_ERROR;
    }

    if (numarg == 3) {
	pc = strtoul(argumentos[2], &end, 0);
	if ((*end != 0) || (end == argumentos[2])) {
	    wsprintf(salidas, "Incorrect pc \"%.50s\"", argumentos[2]);
	    BWCCMessageBox(GetFocus(), salidas, "Error", MB_OK | MB_ICONEXCLAMATION);
	    machPtr->result[0] = '\0';
	    return TCL_ERROR;
	}
    } else {
	pc = 0;
    }
    result = Asm_Assemble(machPtr, (char far *) NULL, argumentos[1], pc, argumentos[1], uno,
	    &size, code);
    if (result != TCL_OK) {
	return result;
    }
    wsprintf(salidas, "Assembler code: %.50s => 0x%lx", argumentos[1], code[0]);
    BWCCMessageBox(GetFocus(), salidas, "Assembler", MB_OK | MB_ICONINFORMATION);
    machPtr->result[0] = '\0';
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Asm_LoadCmd --
 *
 *	Este procedimiento es invocado para procesar el comando
 *	"Cargar Programa".
 *	Ver la documentacion de usuario para conocer mas detalles sobre
 *	este comando.
 *
 * Resultados:
 *	Un resultado de OK o de ERROR.
 *
 * Efectos Laterales:
 *	Ver la documentacion de usuario.
 *
 *----------------------------------------------------------------------
 */

/* Se utilizan argumentos */
long int
Asm_LoadCmd(machPtr, numarg, argumentos)
    DLXV *machPtr;		/* Maquina en cuya memoria se ha de cargar el
				 * programa. */
    int numarg;			/* Contador del numero de ficheros en
				 * argumentos. */
    char argumentos[argus][largus]; /* Array conteniendo los nombres de los
				 * ficheros a ensamblar. */
{
    unsigned long int codeStart, dataStart;
    LoadInfo info;
    long int i;

    /*
     * Usa los valores por defecto para la direccion de comienzo de codigo
     * y de datos.
     */

    codeStart = 0x100;
/* ***    p = Tcl_GetVar(machPtr->interp, "codeStart", 1);
    if (*p != 0) {
	codeStart = strtoul(p, &end, 0);
	if (*end != 0) {
	    sprintf(machPtr->interp->result,
		    "\"codeStart\" variable doesn't contain an address: \"%.50s\"",
		    p);
	    return TCL_ERROR;
	}
    }
*** */
    dataStart = 0x1000;
/* ***    p = Tcl_GetVar(machPtr->interp, "dataStart", 1);
    if (*p != 0) {
	dataStart = strtoul(p, &end, 0);
	if (*end != 0) {
	    sprintf(machPtr->interp->result,
		    "\"dataStart\" variable doesn't contain an address: \"%.50s\"",
		    p);
	    return TCL_ERROR;
	}
    }
*** */
    /*
     * Paso 1: Borra todas la viejas definiciones de simbolos.
     */

    for (i = 1; i < numarg; i++) {
	Sym_DeleteSymbols(machPtr, &argumentos[(int)i]);
    }

    /*
     * Paso 2:  Lee todos los ficheros para construir la tabla de simbolos.
     */

    info.codeAddr = codeStart;
    info.dataAddr = dataStart;
    info.message = NULL;
    info.end = NULL;
    info.totalBytes = 0;
    info.errorCount = 0;
    info.flags = ASM_SIZE_ONLY;
    for (i = 1; i < numarg; i++) {
	ReadFile(argumentos[(int)i], machPtr, &info);
    }

    /*
     * Paso 3: Lee todos los ficheros por segunda vez para ensamblar el codigo.
     */

    info.codeAddr = codeStart;
    info.dataAddr = dataStart;
    info.flags = 0;
    for (i = 1; i < numarg; i++) {
	ReadFile(argumentos[(int)i], machPtr, &info);
	if (info.errorCount > ASM_MAX_ERRORS) {
	    break;
	}
    }

    if (info.message == NULL) {
	return TCL_OK;
    }
    Tcl_Return(machPtr, info.message, TCL_DYNAMIC);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * ReadFile --
 *
 *	Lee un fichero con instrucciones en ensamblador.
 *
 * Resultados:
 *	Ninguno.
 *
 * Efectos Laterales:
 *	La informacion se carga en la memoria *machPtr, e *infoPtr
 *	es modificado (por ejemplo para apuntar a un mensaje de error).
 *
 *----------------------------------------------------------------------
 */

static void
ReadFile(fileName, machPtr, infoPtr)
    char *fileName;		/* Nombre del fichero a leer. */
    DLXV *machPtr;		/* Maquina en cuya memoria se va a cargar
				 * la informacion. */
    register LoadInfo *infoPtr;	/* Informacion acerca del estado del proceso
				 * de ensamblado. */
{
#define MAX_LINE_SIZE 200L
#define MAX_NAME_SIZE 10L
    char line[MAX_LINE_SIZE];
    char pseudoOp[MAX_NAME_SIZE+1];
    FILE *f;
    register char *p;
    long int i, nullTerm;
    char *end, *curToken;
    char savedChar;
    double strtod();

    f = fopen(fileName, "rt");
    if (f == NULL) {
	if (infoPtr->flags & ASM_SIZE_ONLY) {
	    return;
	}
	sprintf((char far *)machPtr->result,
		"Cannot open file \"%.50s\": %.100s", fileName,
		strerror(errno));
	AddErrMsg(machPtr, infoPtr, cero);
	return;
    }

    /*
     * Procesa el fichero, una linea cada vez.
     */

    infoPtr->file = fileName;
    infoPtr->dot = (infoPtr->codeAddr + 3) & ~3;
    infoPtr->flags |= ASM_CODE;
    for (infoPtr->lineNum = 1; ; infoPtr->lineNum++) {
	infoPtr->line = fgets(line, (int)MAX_LINE_SIZE, f);
	if (infoPtr->line == NULL) {
	    if (!feof(f)) {
		sprintf((char far *)machPtr->result, "Cannot read file: %.100s",
			strerror(errno));
		AddErrMsg(machPtr, infoPtr, uno);
	    }
	    break;
	}

	/*
	 * Saltar los espacios en blanco.
	 */

	for (p = line; (*p == ' ') || (*p == '\t'); p++) {
	    /* Bucle con cuerpo vacio:  solo salta espacios. */
	}

	/*
	 * Analiza si existe un simbolo opcional al comienzo de la linea.
	 * Nota:  Se fuerza a que los mensajes de error de simbolos aparezcan
	 * durante la primera pasada, aunque muchos otros errores sean
	 * ignorados durante esta pasada.
	 */

	if (isalpha(*p) || (*p == '_') || (*p == '$')) {
	    curToken = p;
	    for (p++; isalnum(*p) || (*p == '_') || (*p == '$'); p++) {
		/* Bucle con cuerpo vacio:  Para saltar el ultimo simbolo. */
	    }
	    if (*p == ':') {
		*p = 0;
		if (infoPtr->flags & ASM_SIZE_ONLY) {
		    Sym_AddSymbol(machPtr, fileName, curToken, infoPtr->dot, cero);
		    if (*machPtr->result != 0) {
			AddErrMsg(machPtr, infoPtr, uno);
		    }
		}
		*p = ':';
		p++;
	    } else {
		p = curToken;
	    }
	}

	/*
	 * Hay que saltarse las lineas vacias.
	 */

	while ((*p == ' ') || (*p == '\t')) {
	    p++;
	}
	if ((*p == '\n') || (*p == 0)) {
	    continue;
	}

	/*
	 * Si este no es un pseudo operando del ensamblador entonces ensambla
	 * la instruccion y continua.
	 */

	while ((*p == ' ') || (*p == '\t')) {
	    p++;
	}
	if (*p == ';') {
	    continue;
	}
	if (*p != '.') {
	    long int size, code[ASM_MAX_WORDS], result;

	    infoPtr->dot = (infoPtr->dot + 3) & ~3;
	    result = Asm_Assemble(machPtr, fileName, p, infoPtr->dot,
		    line, infoPtr->flags & ASM_SIZE_ONLY, &size, code);
	    if (result == TCL_OK) {
		result = StoreWords(machPtr, infoPtr->dot, code, size);
	    }
	    infoPtr->dot += size*4;
	    goto endOfLine;
	}

	/*
	 * Maneja el caso de que sea un pseudo operando del ensamblador.
	 */

	curToken = p;
	for (i = 0, p++; (i < MAX_NAME_SIZE) && isalpha(*p); i++, p++) {
	    pseudoOp[(unsigned int)i] = *p;
	}
	if (i >= MAX_NAME_SIZE) {
	    IndicateError(machPtr, "Name of operand too long",
		    line, curToken);
	    goto endOfLine;
	}
	pseudoOp[(unsigned int)i] = 0;
	while ((*p == ' ') || (*p == '\t')) {
	    p++;
	}
	if ((pseudoOp[0] == 'a') && (strcmp(pseudoOp, "align") == 0)) {
	    if (Sym_EvalExpr(machPtr, fileName, p, cero, &i, &end) != TCL_OK) {
		IndicateError(machPtr, machPtr->result,
			line, p);
		goto endOfLine;
	    }
	    p = end;
	    if (i == 0) {
		machPtr->result = "\".align 0\" not supported";
		goto endOfLine;
	    } else {
		i = (1 << i) - 1;
		infoPtr->dot = (infoPtr->dot + i) & ~i;
	    }
	} else if ((pseudoOp[0] == 'a') && (strcmp(pseudoOp, "ascii") == 0)) {
	    nullTerm = 0;

	    /*
	     * Lee una o mas cadenas ASCII de la linea de entrada.  Cada
	     * una de ellas deben estar delimitadas por comillas, y entre
	     * ellas debe de haber una separacion por medio de comas.
	     */

	    doString:
	    while (1) {
		while ((*p == ' ') || (*p == '\t')) {
		    p++;
		}
		if (*p != '"') {
		    IndicateError(machPtr,
			    "Error \" at the begin of the string", line, p);
		    goto endOfLine;
		}
		p++;
		i = Gp_PutString(machPtr, p, '"', infoPtr->dot, nullTerm, &end);
		if (*end != '"') {
		    IndicateError(machPtr,
			    "Error \" at the end of the string", line, end-1);
		    goto endOfLine;
		}
		p = end+1;
		infoPtr->dot += i;
		while ((*p == ' ') || (*p == '\t')) {
		    p++;
		}
		if (*p != ',') {
		    break;
		}
		p++;
	    }
	} else if ((pseudoOp[0] == 'a') && (strcmp(pseudoOp, "asciiz") == 0)) {
	    nullTerm = 1;
	    goto doString;
	} else if ((pseudoOp[0] == 'b') && (strcmp(pseudoOp, "byte") == 0)) {
	    while (1) {
		curToken = p;
		if (Sym_EvalExpr(machPtr, fileName, p, cero, &i, &end)
			!= TCL_OK) {
/* *** */	    IndicateError(machPtr, machPtr->result,
			    line, p);
		    goto endOfLine;
		}
		Gp_PutByte(machPtr, infoPtr->dot, i);
		infoPtr->dot += 1;
		for (p = end; (*p == ' ') || (*p == '\t'); p++) {
		    /* Bucle con cuerpo vacio; Se salta los espacios vacios. */
		}
		if (*p != ',') {
		    break;
		}
		p++;
	    }
	} else if ((pseudoOp[0] == 'd') && (strcmp(pseudoOp, "data") == 0)) {
	    if (infoPtr->flags & ASM_CODE) {
		infoPtr->codeAddr = infoPtr->dot;
	    } else {
		infoPtr->dataAddr = infoPtr->dot;
	    }
	    if (Sym_EvalExpr(machPtr, fileName, p, cero, &i, &end) != TCL_OK) {
		Tcl_Return(machPtr, (char far *) NULL, TCL_STATIC);
	    } else {
		p = end;
		infoPtr->dataAddr = i;
	    }
	    infoPtr->dot = infoPtr->dataAddr;
	    infoPtr->flags &= ~ASM_CODE;
	} else if ((pseudoOp[0] == 'g') && (strcmp(pseudoOp, "global") == 0)) {
	    if (!isalpha(*p) && (*p != '_')) {
		IndicateError(machPtr,
			"Name of simbol must start with '_'", line, p);
		goto endOfLine;
	    }
	    curToken = p;
	    while (isalnum(*p) || (*p == '_') || (*p == '$')) {
		p++;
	    }
	    savedChar = *p;
	    *p = 0;
	    if (infoPtr->flags & ASM_SIZE_ONLY) {
		Sym_AddSymbol(machPtr, fileName, curToken, cero,
			SYM_GLOBAL|SYM_NO_ADDR);
		if (*machPtr->result != 0) {
		    AddErrMsg(machPtr, infoPtr, uno);
		}
	    }
	    *p = savedChar;
	} else if ((pseudoOp[0] == 's') && (strcmp(pseudoOp, "space") == 0)) {
	    if (Sym_EvalExpr(machPtr, fileName, p, cero, &i, &end) != TCL_OK) {
		IndicateError(machPtr, machPtr->result,
			line, p);
		goto endOfLine;
	    }
	    p = end;
	    while (i > 0) {
		Gp_PutByte(machPtr, infoPtr->dot, cero);
		infoPtr->dot += 1;
		i -= 1;
	    }
	} else if ((pseudoOp[0] == 't') && (strcmp(pseudoOp, "text") == 0)) {
	    if (infoPtr->flags & ASM_CODE) {
		infoPtr->codeAddr = infoPtr->dot;
	    } else {
		infoPtr->dataAddr = infoPtr->dot;
	    }
	    if (Sym_EvalExpr(machPtr, fileName, p, cero, &i, &end) != TCL_OK) {
		Tcl_Return(machPtr, (char far *) NULL, TCL_STATIC);
	    } else {
		p = end;
		infoPtr->codeAddr = i;
	    }
	    infoPtr->dot = infoPtr->codeAddr;
	    infoPtr->flags |= ASM_CODE;
	} else if ((pseudoOp[0] == 'w') && (strcmp(pseudoOp, "word") == 0)) {
	    while (1) {
		curToken = p;
		if (Sym_EvalExpr(machPtr, fileName, p, cero, &i, &end)
			!= TCL_OK) {
		    IndicateError(machPtr, machPtr->result,
			    line, p);
		    goto endOfLine;
		}
		infoPtr->dot = (infoPtr->dot + 3) & ~3;
		(void) StoreWords(machPtr, infoPtr->dot, &i, uno);
		infoPtr->dot += 4;
		for (p = end; (*p == ' ') || (*p == '\t'); p++) {
		    /* Bucle con cuerpo vacio;  Salta los espacios. */
		}
		if (*p != ',') {
		    break;
		}
		p++;
	    }
	} else if ((pseudoOp[0] == 'f') && (strcmp(pseudoOp, "float") == 0)) {
	    while (1) {
		float f;
		long int *pi = (long int *)&f;

		curToken = p;
		f = strtod(p, &end);
		if (p == end) {
		    IndicateError(machPtr, "Simple precision number illegal", line, p);
		    goto endOfLine;
		}
		infoPtr->dot = (infoPtr->dot + 3) & ~3;
		(void) StoreWords(machPtr, infoPtr->dot, pi, uno);
		infoPtr->dot += 4;
		for (p = end; (*p == ' ') || (*p == '\t'); p++) {
		    /* Cuerpo vacio;  Salta los espacios. */
		}
		if (*p != ',') {
		    break;
		}
		p++;
	    }
	} else if ((pseudoOp[0] == 'd') && (strcmp(pseudoOp, "double") == 0)) {
	    while (1) {
		double d;
		long int *pi = (long int *)&d;

		curToken = p;
		d = strtod(p, &end);
		if (p == end) {
		    IndicateError(machPtr, "Double precision number illegal", line, p);
		    goto endOfLine;
		}
		infoPtr->dot = (infoPtr->dot + 3) & ~3;
		(void) StoreWords(machPtr, infoPtr->dot, pi, dos);
		infoPtr->dot += 8;
		for (p = end; (*p == ' ') || (*p == '\t'); p++) {
		    /* Cuerpo vacio;  Salta los espacios. */
		}
		if (*p != ',') {
		    break;
		}
		p++;
	    }
	} else {
	    IndicateError(machPtr, "Operand not recognized", line,
		    curToken);
	    goto endOfLine;
	}

	/*
	 * Comprueba si existen caracteres extraos al final de la linea.
	 */

	while (isspace(*p)) {
	    p++;
	}
	if ((*p != '#') && (*p != 0)) {
	    IndicateError(machPtr, "Characters after the end of line",
		    line, p);
	}

	/*
	 * Se ha procesado la linea.  Si ha habido un error, se aade a la
	 * lista de mensajes de error que se ha ido acumulando durante el
	 * ensamblado. Se incrementa la cantidad de memoria alojada para
	 * los mensajes de error si ello es necesario para incluir el nuevo
	 * mensaje.
	 */

	endOfLine:
	if (*machPtr->result != 0) {
	    if (infoPtr->flags & ASM_SIZE_ONLY) {
		Tcl_Return(machPtr, (char far *) NULL, TCL_STATIC);
	    } else {
		AddErrMsg(machPtr, infoPtr, uno);
		if (infoPtr->errorCount > ASM_MAX_ERRORS) {
		    goto endOfFile;
		}
	    }
	}
    }

    endOfFile:
    fclose(f);
    if (infoPtr->flags & ASM_CODE) {
	infoPtr->codeAddr = infoPtr->dot;
    } else {
	infoPtr->dataAddr = infoPtr->dot;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * AddErrMsg --
 *
 *	Dado un mensaje de error almacenado en la maquina, lo aade a la
 *	lista de mensajes de error que se han ido acumulando durante el
 *	ensamblado. Ademas inicializa (pone un string vacio) el mensaje
 *	que ha sido pasado en la maquina.
 *
 * Resultados:
 *	Ninguno.
 *
 * Efectos Laterales:
 *	El mensaje es aadido a la lista de mensajes en infoPtr, y el
 *	campo machPtr->result es reinicializado.
 *
 *----------------------------------------------------------------------
 */

static void
AddErrMsg(maq, infoPtr, addHeader)
    DLXV *maq;			/* Maquina conteniendo el mensaje de error. */
    register LoadInfo *infoPtr;	/* Estado del ensamblado, al cual se aaden
				 * los nuevos mensajes de error. */
    long int addHeader;		/* No-cero significa aadir en la cabecera del
				 * mensaje la identificacion del fichero y el
				 * numero de linea donde se ha producido el
				 * error. */
{
    long int length, hdrLength, totalLength;
    char header[100];

    length = _fstrlen((char far *)maq->result);
    if (length == 0) {
	return;
    }
    if (addHeader) {
	sprintf(header, "%.50s(%ld): ", infoPtr->file,
		infoPtr->lineNum);
    } else {
	header[0] = 0;
    }
    hdrLength = _fstrlen(header);
    totalLength = hdrLength + length + 2;

    /*
     * Aumenta el area de mensajes de error si el area actual no es lo
     * suficientemente grande.
     */

    if (totalLength > ((infoPtr->message + infoPtr->totalBytes)
	    - (infoPtr->end + 1))) {
	char huge *newMsg;

	if (infoPtr->totalBytes == 0) {
	    infoPtr->totalBytes = 4*totalLength;
	} else {
	    infoPtr->totalBytes = 2*(infoPtr->totalBytes + totalLength);
	}
	newMsg = farcalloc(1, (unsigned long int) infoPtr->totalBytes);
	if (infoPtr->message != NULL) {
	    strcpy((char far *)newMsg, (char far *)infoPtr->message);
	    infoPtr->end += newMsg - infoPtr->message;
	} else {
	    infoPtr->end = newMsg;
	}
	infoPtr->message = newMsg;
    }
    if (infoPtr->end != infoPtr->message) {
	*infoPtr->end = '\n';
	infoPtr->end += 1;
    }
    sprintf((char far *)infoPtr->end, "%s%s", header, (char far *)maq->result);
    infoPtr->end += hdrLength + length;
    infoPtr->errorCount += 1;

    Tcl_Return(maq, (char far *) NULL, TCL_STATIC);
}

/*
 *----------------------------------------------------------------------
 *
 * StoreWords --
 *
 *	Almacena un determinado numero de palabras en la memoria de la
 *	maquina.
 *
 * Resultados:
 *	Resultado de OK (mas un string vacio); o resultado ERROR (mas
 *	informacion de error retornada a traves de machPtr->result.
 *
 * Efectos Laterales:
 *	La memoria de MachPtr es modificada al almacenar la nueva
 *	informacion.
 *
 *----------------------------------------------------------------------
 */

static long int
StoreWords(machPtr, address, wordPtr, numWords)
    register DLXV *machPtr;		/* Maquina en la cual almacenar. */
    unsigned long int address;		/* Direccion de la memoria de la
					 * maquina. */
    long int *wordPtr;			/* Palabras que se almacenaran en
					 * memoria de la maquina. */
    long int numWords;			/* Numero de palabras a almacenar. */
{
    long int index;
    register MemWord huge *memPtr;

    for ( ; numWords > 0; wordPtr++, address += 4, numWords--) {
	index = ADDR_TO_INDEX(address);
	if (index >= machPtr->numWords) {
	    sprintf((char far *)machPtr->result,
		    "Cannot store 0x%lx: unexisting position",
		    address);
	    return TCL_ERROR;
	}
	memPtr = machPtr->memPtr + index;
	memPtr->value = *wordPtr;
	memPtr->opCode = OP_NOT_COMPILED;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * IndicateError --
 *
 *	Genera un mensaje de error que indica la posicion del string
 *	donde se ha detectado el error.
 *
 * Resultados:
 *	No hay valor de retorno.  machPtr->result es modificado para contener
 *	el errMsg seguido por el string, con la posicion pos indicada en el
 *	string.
 *
 * Efectos Laterales:
 *	Ninguno.
 *
 *----------------------------------------------------------------------
 */
#pragma argsused
static void
IndicateError(maq, errMsg, string, pos)
    DLXV *maq; 		/* Maquina en la que almacenar el mensaje de error.
			 * El area que se utiliza debe estar en el estado
			 * inicial de cadena vacia. */
    char *errMsg;	/* Mensaje describiendo el problema. */
    char *string;	/* Cadena de entrada que contiene el problema. */
    char *pos;		/* Posicion en el string del caracter donde se detecto
			 * el problema. */
{
    long int msgLength, stringLength;
    char *newMsg;

    msgLength = _fstrlen(errMsg);
    stringLength = _fstrlen(string);
     if (string[(unsigned int)(stringLength-1)] == '\n') {
	string[(unsigned int)(stringLength-1)]='\x0';
	}

    /*
     * Siempre asigna nueva zona de almacenamiento para el nuevo mensaje.
     * Esto es necesario debido a: (a) El espacio requerido puede exceder
     * el tamao del buffer estatico de resultados, y (b) "errMsg" puede
     * estar actualmente en el buffer estatico por lo tanto debemos tener
     * cuidado de no destruirlo mientras generamos el nuevo mensaje.
     */

    newMsg = farcalloc(1, (unsigned long int) (msgLength + stringLength + 10));
    sprintf(newMsg, "%s => %s", errMsg, string);
    Tcl_Return(maq, newMsg, TCL_DYNAMIC);
}
